"""Separate designated agent ids

Revision ID: 4db74efaf82e
Revises: fb01c44db0d6
Create Date: 2025-09-03 12:52:38.558284

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4db74efaf82e"
down_revision = "fb01c44db0d6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "sessions", sa.Column("designated_agent_ids", sa.ARRAY(sa.String()), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("sessions", "designated_agent_ids")
    # ### end Alembic commands ###
