"""add kernel last seen

Revision ID: 42998e8c8bb4
Revises: c7fedaca4883
Create Date: 2025-05-13 15:42:29.976150

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "42998e8c8bb4"
down_revision = "c7fedaca4883"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column(
            "last_seen", sa.DateTime(timezone=True), server_default=sa.text("NULL"), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "last_seen")
    # ### end Alembic commands ###
