"""Base action class for app_config service."""

from __future__ import annotations

from ai.backend.manager.actions.action import BaseAction


class AppConfigAction(BaseAction):
    """Base class for all app_config actions."""
