"""Add verification_result to artifact_revisions

Revision ID: 4cfe35b1e60a
Revises: b55309328455
Create Date: 2025-11-08 01:53:01.604948

"""

import sqlalchemy as sa
from alembic import op

revision = "4cfe35b1e60a"
down_revision = "b55309328455"

branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "artifact_revisions",
        sa.Column("verification_result", sa.JSON(none_as_null=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("artifact_revisions", "verification_result")
    # ### end Alembic commands ###
